//------------------------------ $Keywords ----------------------------------
// SelfImage - Disk image manipulation program
// SelfImage_DecompressionThread.cpp - Data decompression thread class
// Copyright 2005, Kurt Fitzner <kfitzner@excelcia.org>
//---------------------------------------------------------------------------
// This file is part of SelfImage.
//
// SelfImage is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License (Version 2) as
// published by the Free Software Foundation.
//
// SelfImage is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// VCS: $Version: 0 $ $Revision: 2 $
/*
$History: **** V 0.1 by kfitzner ****
$History: * selfimage_decompressionthread.cpp - 2005-11-27 6:08:37 AM - 14280 Bytes
$History: * selfimage_decompressionthread.h - 2005-11-27 4:56:56 AM - 2539 Bytes
$History: * Initial Check-in
$History: **** Latest ** V 0.2 by kfitzner ** 2005-11-28 1:39:55 PM ****
$History: * Set FFinished to true at end of loops
*/
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// File Notes:
//---------------------------------------------------------------------------
// 26 Nov 2005 - Kurt Fitzner <kfitzner@excelcia.org>
//
// Decompression thread.  For each type of compression supported, this
// thread has to support decompressing in two ways.  The block method, where
// the input is a compressed file that's divided up into a blocks of
// discreetly compressed data, and the stream method, where the input file
// is one big compressed stream.  Detection of the type is done way back in
// the TImageStore class, and that detection is used to determine how many
// copies of this thread are started.  Block decompression can be done in
// parallel - stream can only be done linearly.  This thread also doesn't
// have to worry about dividing up the blocks - that is also done in
// TImageStore.  But, we do need to handle the decompression very differently
// for each type.
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "SelfImage_DecompressionThread.h"
#include "zlib.h"
#include "bzlib.h"
#include "SelfImage_Exceptions.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Constructor
//
__fastcall TSelfImageDecompressionThread::TSelfImageDecompressionThread(TCompressionFormat CompressionFormat, TImageBuffer *InputBuffer, TImageBuffer *OutputBuffer): TSelfImageThread()
{
  this->InputBuffer  = InputBuffer;
  this->OutputBuffer = OutputBuffer;
  this->CompressionFormat = CompressionFormat;
  if (CompressionFormat == cfGzipBlock || CompressionFormat == cfBzip2Block) {
    BlockDecompress = true;
    InputBuffer->AllocationSister = OutputBuffer;
  } else
    BlockDecompress = false;
  BytesRead = 0;
  BytesWritten = 0;
}  // __fastcall TSelfImageCompressionThread::TSelfImageCompressionThread(TCompressionMethod CompressionMethod, TImageBuffer *InputBuffer, TImageBuffer *OutputBuffer): TSelfImageThread()
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Function to set the thread's name for debugging purposes.  This is done
// by filling in a special structure and firing off an exception.  A
// compatible debugger knows to look in exceptions of this type and look for
// the thread's name.  A little bit of an odd way of doing things, but it
// seems to work.
//
void TSelfImageDecompressionThread::SetName()
{
  THREADNAME_INFO info;
  info.dwType = 0x1000;
  info.szName = "DecompressionThread";
  info.dwThreadID = -1;
  info.dwFlags = 0;

  __try {
    RaiseException( 0x406D1388, 0, sizeof(info)/sizeof(DWORD),(DWORD*)&info );
  }
  __except (EXCEPTION_CONTINUE_EXECUTION) {}
}  // void TSelfImageCompressionThread::SetName()
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// We have three different main loops - one for block decompression, the
// other two for stream decompression.  All we do here is decide which loop
// to use and pass control.
//
void __fastcall TSelfImageDecompressionThread::Execute()
{
  SetName();

  if (BlockDecompress)
    BlockDecompressLoop();
  else
    StreamDecompressLoop();
    
}  // void __fastcall TSelfImageDecompressionThread::Execute()
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Block decompress loop - decompress files that were compressed in blocks
//
void __fastcall TSelfImageDecompressionThread::BlockDecompressLoop(void)
{
  bool bEOF = false;

  try {
    while (!bEOF && !Terminated) {
      TBufferChunk *ReadChunk, *DecompressChunk;
      ReadChunk = InputBuffer->GetChunk(csAllocRead, &Terminated);
      if (!ReadChunk)
        continue;
      DecompressChunk = ReadChunk->Sister;
      // If we didn't get a sister chunk and the thread isn't terminating, then there is a problem.
      if (!DecompressChunk)
        if (!Terminated)
          throw ESelfImageSystemError("Buffer allocation failed to return both a read and write buffer chunk.  This is probably a programming error - please report it to the maintainers.");
        else
          continue;
      bEOF = ReadChunk->EOF;
      DecompressChunk->EOF = bEOF;
      // Only compress if we actually have data - otherwise just set the output chunk's size to zero.
      if (ReadChunk->Size) {
        DecompressBlock(ReadChunk, DecompressChunk);
        BytesRead += ReadChunk->Size;
        BytesWritten += DecompressChunk->Size;
      } else {
        DecompressChunk->Size = 0;
      }  // else if (!ReadChunk->Size)
      InputBuffer->ReleaseChunk(ReadChunk);
    }  // while (!bEOF && !Terminated)
    FFinished = true;
  }  // try
  // C++ Builder doesn't support passing exceptions from a thread.  So, we have to catch the exception here and
  // manually pass word that an error has occured.  This is done with the ErrorFlag property.
  catch (Exception &e) {
    FErrorFlag = true;
    FErrorMessage = "Decompression thread encountered " + e.ClassName() + " exception: \"" + e.Message + "\"";
    FFinished = true;
  }  // catch
}  // void __fastcall TSelfImageDecompressionThread::BlockDecompressLoop(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Decompress one block
//
void __fastcall TSelfImageDecompressionThread::DecompressBlock(TBufferChunk *Source, TBufferChunk *Target)
{
  switch (CompressionFormat) {
    // Gzip-compressed file that is divided up into individual blocks.
    // Each block is a complete mini-gzip stream in and of itself.
    case cfGzipBlock: {
      z_stream *zsStream = NULL;

      try {
        int ret;
        unsigned char *InData = (unsigned char *)Source->Data;
        unsigned char *OutData = (unsigned char *)Target->Data;

        // Configure the z_stream - windowbits > 15 means to accept a gzip stream, not just a zlib stream
        zsStream = new z_stream;
        memset(zsStream, 0, sizeof(z_stream));
        zsStream->next_in  = InData;
        zsStream->avail_in = Source->Size;
        ret = inflateInit2(zsStream, 31);
        if (ret != Z_OK)
          throw ESelfImageCompressionError("Initialization of a gzip decompression block failed.");

        zsStream->avail_out = Target->MaxSize;
        zsStream->next_out  = OutData;
        do ret = inflate(zsStream, Z_FINISH); while (ret == Z_OK);
        if (ret != Z_STREAM_END)
          throw ESelfImageCompressionError("Failed to decompress a block - " + String(zsStream->msg));

        ret = inflateEnd(zsStream);
        if (ret != Z_OK)
          throw ESelfImageCompressionError("Failed to clean up after decompressing a block.");

        Target->Size = zsStream->total_out;
      }  // try
      __finally {
        if (zsStream)
          delete zsStream;
      }  // __finally
    }  // case cfGzipBlock:
    break;

    case cfBzip2Block: {
      int ret;
      unsigned char *InData = (unsigned char *)Source->Data;
      unsigned char *OutData = (unsigned char *)Target->Data;
      unsigned int BlockSize;

      BlockSize = Target->MaxSize;
      BZ2_bzBuffToBuffDecompress(OutData, &BlockSize, InData, Source->Size, false, 0);
      Target->Size = BlockSize;
    }  // case cgBzip2Block:
    break;
  }  // switch (CompressionFormat)
}  // void TSelfImageDecompressionThread::Decompress(TBufferChunk *Source, TBufferChunk *Target)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// gzip stream decompress loop - decompress files that were compressed
// normally - ie: aren't divided up into little "streamlet" blocks.
//
void __fastcall TSelfImageDecompressionThread::StreamDecompressLoop(void)
{
  bool bEOF = false;

  try {
    while (!bEOF && !Terminated) {
      TBufferChunk *ReadChunk;
      ReadChunk = InputBuffer->GetChunk(csAllocRead, &Terminated);
      if (!ReadChunk)
        continue;
      // Only compress if we actually have data - otherwise just set the output chunk's size to zero.
      if (ReadChunk->Size) {
        DecompressStream(ReadChunk);
        BytesRead += ReadChunk->Size;
      }  // else if (!ReadChunk->Size)
      bEOF = ReadChunk->EOF;
      InputBuffer->ReleaseChunk(ReadChunk);
    }  // while (!bEOF && !Terminated)
    FFinished = true;
  }  // try
  // C++ Builder doesn't support passing exceptions from a thread.  So, we have to catch the exception here and
  // manually pass word that an error has occured.  This is done with the ErrorFlag property.
  catch (Exception &e) {
    FErrorFlag = true;
    FErrorMessage = "Decompression thread encountered " + e.ClassName() + " exception: \"" + e.Message + "\"";
    FFinished = true;
  }  // catch

}  // void __fastcall TSelfImageDecompressionThread::StreamDecompressLoop(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Decompress part of a stream - one input block goes out to one or more
// output blocks.  When using the bzip2 "low level interface", it is almost
// identical to using zlib.
//
void __fastcall TSelfImageDecompressionThread::DecompressStream(TBufferChunk *Source)
{
  static bool bFirst = true;
  int ret;
  unsigned char *InData = (unsigned char *)Source->Data;
  unsigned char *OutData;
  TBufferChunk *Target;

  switch (CompressionFormat) {
    case cfGzipStream: {
      static z_stream zsStream;

      if (bFirst) {
        bFirst = false;
        ret = inflateInit2(&zsStream, 31);
        if (ret != Z_OK)
          throw ESelfImageCompressionError("Initialization of gzip decompression failed.");
      }  // if (bFirst)

      zsStream.next_in = InData;
      zsStream.avail_in = Source->Size;
      do {
        Target = OutputBuffer->GetChunk(csAllocWrite, &Terminated);
        if (!Target)
          return;
        zsStream.next_out = (unsigned char *)Target->Data;
        zsStream.avail_out = Target->MaxSize;
        ret = inflate(&zsStream, Z_NO_FLUSH);
        if (ret == Z_STREAM_END) {
          ret = inflateEnd(&zsStream);
          if (ret != Z_OK)
            throw ESelfImageCompressionError("Failed to clean up after decompressing.");
          Target->EOF = true;
        }  // if (ret == Z_STREAM_END)
        else if (ret != Z_OK)
          throw ESelfImageCompressionError("Error while decompressing - " + String(zsStream.msg));
        Target->Size = Target->MaxSize - zsStream.avail_out;
        BytesWritten += Target->Size;
        OutputBuffer->ReleaseChunk(Target);
      } while (zsStream.avail_in && ret != Z_STREAM_END);
    }  // case cfGzipStream:

    case cfBzip2Stream: {
      static bz_stream bzsStream;

      if (bFirst) {
        bFirst = false;
        ret = BZ2_bzDecompressInit(&bzsStream, 0, false);
        if (ret != BZ_OK)
          throw ESelfImageCompressionError("Initialization of bzip2 decompression failed.");
      }  // if (bFirst)

      bzsStream.next_in = InData;
      bzsStream.avail_in = Source->Size;
      do {
        Target = OutputBuffer->GetChunk(csAllocWrite, &Terminated);
        if (!Target)
          return;
        bzsStream.next_out = (unsigned char *)Target->Data;
        bzsStream.avail_out = Target->MaxSize;
        ret = BZ2_bzDecompress(&bzsStream);
        if (ret == BZ_STREAM_END) {
          ret = BZ2_bzDecompressEnd(&bzsStream);
          if (ret != BZ_OK)
            throw ESelfImageCompressionError("Failed to clean up after decompressing.");
          Target->EOF = true;
        }  // if (ret == Z_STREAM_END)
        else if (ret != BZ_OK)
          throw ESelfImageCompressionError("Bzip2 error while decompressing.");
        Target->Size = Target->MaxSize - bzsStream.avail_out;
        BytesWritten += Target->Size;
        OutputBuffer->ReleaseChunk(Target);
      } while (bzsStream.avail_in && ret != BZ_STREAM_END);
    }  // case cfBzip2Stream:

  }  // switch (CompressionFormat)
}  // void __fastcall TSelfImageDecompressionThread::DecompressStream(TBufferChunk *Source, TImageBuffer OutputBuffer)
//---------------------------------------------------------------------------

